/* slurp-resource.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-resource"

#include "slurp-resource.h"

// TODO: Auto set updated using g_get_monotonic_time ()
typedef struct
{
    gchar *origin;
    gchar *destination;
    gboolean overwrite;

    GFile *file;
    gchar buffer[16384]; // 16 * 1024
    GOutputStream *output;

    guint64 total;
    guint64 downloaded;
    guint64 updated;
} SlurpResourcePrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpResource, slurp_resource, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_ORIGIN,
    PROP_DESTINATION,
    PROP_OVERWRITE,

    PROP_FILE,
    PROP_BUFFER,
    PROP_OUTPUT,

    PROP_TOTAL,
    PROP_DOWNLOADED,
    PROP_UPDATED,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    STARTED,
    FINISHED,
    CANCELLED,
    PROGRESS,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

enum {
    SLURP_RESOURCE_ERROR,
    SLURP_RESOURCE_ERROR_INVALID,
    SLURP_RESOURCE_ERROR_ORIGIN,
    SLURP_RESOURCE_ERROR_ORIGIN_MISSING,
    SLURP_RESOURCE_ERROR_ORIGIN_INVALID_URI,
    SLURP_RESOURCE_ERROR_ORIGIN_INVALID_PROTOCOL,
    SLURP_RESOURCE_ERROR_DESTINATION,
    SLURP_RESOURCE_ERROR_DESTINATION_MISSING,
    SLURP_RESOURCE_ERROR_DESTINATION_INVALID_URI,
    SLURP_RESOURCE_ERROR_WRITABLE
};

gboolean
slurp_resource_is_valid_origin (SlurpResource  *self,
                                GCancellable   *cancellable,
                                GError        **error)
{
/*
    g_set_error (error,
                 SLURP_RESOURCE_ERROR,
                 SLURP_RESOURCE_ERROR_ORIGIN,
                 "Origin is invalid");
*/

    //g_file_get_uri_scheme ()

    return TRUE;
}

gboolean
slurp_resource_is_valid_destination (SlurpResource  *self,
                                     GCancellable   *cancellable,
                                     GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (!priv->file)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_DESTINATION_MISSING,
                     "Destination is not set");
        return FALSE;
    }

    return TRUE;
}

gboolean
slurp_resource_is_writable (SlurpResource  *self,
                            GCancellable   *cancellable,
                            GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (!priv->file)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_DESTINATION_MISSING,
                     "Destination is not set");
        return FALSE;
    }

    if (!priv->overwrite && g_file_query_exists (priv->file, cancellable))
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_WRITABLE,
                     "Destination is not writable");
        return FALSE;
    }

    return TRUE;
}

gboolean
slurp_resource_is_valid (SlurpResource  *self,
                         GCancellable   *cancellable,
                         GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    GError *internal_error = NULL;

    if (!slurp_resource_is_valid_origin (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    if (!slurp_resource_is_valid_destination (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    if (!slurp_resource_is_writable (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    return TRUE;

out:
    g_clear_error (&internal_error);
    return FALSE;
}


static void
slurp_resource_finalize (GObject *object)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_clear_pointer (&priv->origin, g_free);
    g_clear_pointer (&priv->destination, g_free);
    g_clear_pointer (&priv->file, g_object_unref);
    g_clear_pointer (&priv->output, g_object_unref);

    G_OBJECT_CLASS (slurp_resource_parent_class)->finalize (object);
}

static void
slurp_resource_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_ORIGIN:
            g_value_set_string (value, priv->origin);
            break;
        case PROP_DESTINATION:
            g_value_set_string (value, priv->destination);
            break;

        case PROP_OVERWRITE:
            g_value_set_boolean (value, priv->overwrite);
            break;

        case PROP_FILE:
            g_value_set_object (value, priv->file);
            break;
/*
        case PROP_BUFFER:
            g_value_set_string (value, priv->buffer);
            break;

        case PROP_OUTPUT:
            g_value_set_object (value, priv->output);
            break;
*/
        case PROP_TOTAL:
            g_value_set_uint (value, priv->total);
            break;

        case PROP_DOWNLOADED:
            g_value_set_uint (value, priv->downloaded);
            break;

        case PROP_UPDATED:
            g_value_set_uint (value, priv->updated);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_ORIGIN:
            priv->origin = g_value_dup_string (value);
            break;

        case PROP_DESTINATION:
            priv->destination = g_value_dup_string (value);
            break;

        case PROP_OVERWRITE:
            priv->overwrite = g_value_get_boolean (value);
            break;

        case PROP_FILE:
            priv->file = g_value_dup_object (value);
            break;

/*
        case PROP_BUFFER:
            priv->buffer = g_value_dup_string (value);
            break;

        case PROP_OUTPUT:
            priv->output = g_value_dup_object (value);
            break;
*/
        case PROP_TOTAL:
            priv->total = g_value_get_uint (value);
            break;

        case PROP_DOWNLOADED:
            priv->downloaded = g_value_get_uint (value);
            break;

        case PROP_UPDATED:
            priv->updated = g_value_get_uint (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_class_init (SlurpResourceClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_resource_finalize;
    object_class->get_property = slurp_resource_get_property;
    object_class->set_property = slurp_resource_set_property;

	/**
	 * SlurpResource:origin:
	 *
	 * The original resource to download from in the form of a URI.
	 */
    properties[PROP_ORIGIN] =
        g_param_spec_string ("origin",
                             "Original resource URI",
                             "The original resource to download from in the form of a URI.",
                             NULL,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * SlurpResource:destination:
	 *
	 * The destination which the resource will be saved to in the form of a URI
	 */
    properties[PROP_DESTINATION] =
        g_param_spec_string ("destination",
                             "Destination resource URI",
                             "The destination which the resource will be saved to in the form of a URI.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * SlurpResource:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);


    properties[PROP_FILE] =
        g_param_spec_object ("file",
                             "a GFile",
                             "A GFile used to write the resource to the system.",
                             G_TYPE_FILE,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    properties[PROP_BUFFER] =
        g_param_spec_string ("buffer",
                             "",
                             "",
                             NULL,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    properties[PROP_OUTPUT] =
        g_param_spec_object ("output",
                             "a GOutput",
                             "A GOutputStream used top read the resource.",
                             G_TYPE_OUTPUT_STREAM,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    properties[PROP_TOTAL] =
        g_param_spec_uint ("total",
                           "Total bytes",
                           "The total number of bytes of the resource.",
                           0,
                           -1,
                           0,
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    properties[PROP_DOWNLOADED] =
        g_param_spec_uint ("downloaded",
                           "Total downloaded bytes",
                           "The total number of downloaded byte of the resource.",
                           0,
                           -1,
                           0,
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    /* g_get_monotonic_time() */
    properties[PROP_UPDATED] =
        g_param_spec_uint ("updated",
                           "Last time progress was updated",
                           "The last time the progress of the resource download was updated using the system monotonic time.",
                           0,
                           -1,
                           0,
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);


    properties[PROP_DESTINATION] =
        g_param_spec_string ("destination",
                             "Destination resource URI",
                             "The destination which the resource will be saved to in the form of a URI.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

    signals[STARTED] =
        g_signal_new ("started",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);
    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    signals[PROGRESS] =
        g_signal_new ("progress",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      3,
                      G_TYPE_UINT64,
                      G_TYPE_UINT64,
                      G_TYPE_UINT64);
}

static void
slurp_resource_init (SlurpResource *self)
{
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpResource *
slurp_resouce_new (void)
{
    return g_object_new (SLURP_TYPE_RESOURCE, NULL);
}
